@@ -2,6 +2,7 @@
 
 from errno import EINTR
 import os
+import fcntl
 import select
 import signal
 import sys
@@ -39,9 +40,10 @@ class Manager(object):
         self.askpass = opts.askpass
         self.outdir = opts.outdir
         self.errdir = opts.errdir
-        self.iomap = IOMap()
+        self.iomap = make_iomap()
 
-        self.taskcount = 0
+        self.next_nodenum = 0
+        self.numnodes = 0
         self.tasks = []
         self.running = []
         self.done = []
@@ -95,9 +97,6 @@ class Manager(object):
         signal.signal(signal.SIGCHLD, signal.SIG_DFL)
 
     def set_sigchld_handler(self):
-        # TODO: find out whether set_wakeup_fd still works if the default
-        # signal handler is used (I'm pretty sure it doesn't work if the
-        # signal is ignored).
         signal.signal(signal.SIGCHLD, self.handle_sigchld)
         # This should keep reads and writes from getting EINTR.
         if hasattr(signal, 'siginterrupt'):
@@ -105,20 +104,17 @@ class Manager(object):
 
     def handle_sigchld(self, number, frame):
         """Apparently we need a sigchld handler to make set_wakeup_fd work."""
-        # Write to the signal pipe (only for Python <2.5, where the
-        # set_wakeup_fd method doesn't exist).
-        if self.iomap.wakeup_writefd:
-            os.write(self.iomap.wakeup_writefd, '\0')
         for task in self.running:
             if task.proc:
                 task.proc.poll()
-        # Apparently some UNIX systems automatically resent the SIGCHLD
+        # Apparently some UNIX systems automatically reset the SIGCHLD
         # handler to SIG_DFL.  Reset it just in case.
         self.set_sigchld_handler()
 
     def add_task(self, task):
         """Adds a Task to be processed with run()."""
         self.tasks.append(task)
+        self.numnodes += 1
 
     def update_tasks(self, writer):
         """Reaps tasks and starts as many new ones as allowed."""
@@ -143,8 +139,8 @@ class Manager(object):
         while 0 < len(self.tasks) and len(self.running) < self.limit:
             task = self.tasks.pop(0)
             self.running.append(task)
-            task.start(self.taskcount, self.iomap, writer, self.askpass_socket)
-            self.taskcount += 1
+            task.start(self.next_nodenum, self.numnodes, self.iomap, writer, self.askpass_socket)
+            self.next_nodenum += 1
 
     def reap_tasks(self):
         """Checks to see if any tasks have terminated.
@@ -209,13 +205,9 @@ class IOMap(object):
 
         # Setup the wakeup file descriptor to avoid hanging on lost signals.
         wakeup_readfd, wakeup_writefd = os.pipe()
+        fcntl.fcntl(wakeup_writefd, fcntl.F_SETFL, os.O_NONBLOCK)
         self.register_read(wakeup_readfd, self.wakeup_handler)
-        # TODO: remove test when we stop supporting Python <2.5
-        if hasattr(signal, 'set_wakeup_fd'):
-            signal.set_wakeup_fd(wakeup_writefd)
-            self.wakeup_writefd = None
-        else:
-            self.wakeup_writefd = wakeup_writefd
+        signal.set_wakeup_fd(wakeup_writefd)
 
     def register_read(self, fd, handler):
         """Registers an IO handler for a file descriptor for reading."""
@@ -270,6 +262,65 @@ class IOMap(object):
                 raise FatalError
 
 
+class PollIOMap(IOMap):
+    """A manager for file descriptors and their associated handlers.
+
+    The poll method dispatches events to the appropriate handlers.
+    Note that `select.poll` is not available on all operating systems.
+    """
+    def __init__(self):
+        self._poller = select.poll()
+        super(PollIOMap, self).__init__()
+
+    def register_read(self, fd, handler):
+        """Registers an IO handler for a file descriptor for reading."""
+        super(PollIOMap, self).register_read(fd, handler)
+        self._poller.register(fd, select.POLLIN)
+
+    def register_write(self, fd, handler):
+        """Registers an IO handler for a file descriptor for writing."""
+        super(PollIOMap, self).register_write(fd, handler)
+        self._poller.register(fd, select.POLLOUT)
+
+    def unregister(self, fd):
+        """Unregisters the given file descriptor."""
+        super(PollIOMap, self).unregister(fd)
+        self._poller.unregister(fd)
+
+    def poll(self, timeout=None):
+        """Performs a poll and dispatches the resulting events."""
+        if not self.readmap and not self.writemap:
+            return
+        try:
+            event_list = self._poller.poll(timeout)
+        except select.error:
+            _, e, _ = sys.exc_info()
+            errno = e.args[0]
+            if errno == EINTR:
+                return
+            else:
+                raise
+        for fd, event in event_list:
+            if event & (select.POLLIN | select.POLLHUP):
+                handler = self.readmap[fd]
+                handler(fd, self)
+            if event & (select.POLLOUT | select.POLLERR):
+                handler = self.writemap[fd]
+                handler(fd, self)
+
+
+def make_iomap():
+    """Return a new IOMap or PollIOMap as appropriate.
+
+    Since `select.poll` is not implemented on all platforms, this ensures that
+    the most appropriate implementation is used.
+    """
+    if hasattr(select, 'poll'):
+        return PollIOMap()
+    else:
+        return IOMap()
+
+
 class Writer(threading.Thread):
     """Thread that writes to files by processing requests from a Queue.
 
@@ -299,14 +350,19 @@ class Writer(threading.Thread):
                 return
 
             if data == self.OPEN:
-                self.files[filename] = open(filename, 'wb', buffering=1)
-                psshutil.set_cloexec(self.files[filename])
+                self.files[filename] = None
             else:
                 dest = self.files[filename]
                 if data == self.EOF:
-                    dest.close()
+                    if dest is not None:
+                        dest.close()
                 else:
+                    if dest is None:
+                        dest = self.files[filename] = open(
+                            filename, 'wb', buffering=1)
+                        psshutil.set_cloexec(dest)
                     dest.write(data)
+                    dest.flush()
 
     def open_files(self, host):
         """Called from another thread to create files for stdout and stderr.
